from .utils import get_module
""" Module containing Samples class """

class Sample():
    """
    Class describing the used sample by its config and the used sequence. 
    """

    _master_config_path = None
    master_config = None

    def __init__(
            self, name: str,
            opx_config: dict,
            divider_config: dict,
            param_config = None,):
        """
        Constructor class for 'Sample' class.

        Args:
            name (str): Name of the used sample
            config (dict): OPX configuration file for sample
            self.elements (list): List of all quantum elements
        """
        self.name = name
        self.config = opx_config
        self.param_config = param_config
        self.divider_config = divider_config
        self.elements = list(self.config['elements'].keys())

    @property
    def master_config_path(self):
        """
        Getter for master_config.

        Returns:
            str: The current configuration dictionary.
        """
        return self._master_config_path

    @master_config_path.setter
    def master_config_path(self, config_path):
        """
        Setter for master_config_path.
        Validates and sets the configuration path as a string.

        Args:
            config_path (str): A dictionary containing configuration data.

        Raises:
            ValueError: If the config is not a string.
        """
        if not isinstance(config_path, str):
            raise ValueError("master_config_path must be a str.")
        self._master_config_path = config_path
        mc = get_module('mc', config_path)
        if not hasattr(mc, 'config'):
            raise AttributeError(f"Dictionary 'config' not found in the file {self._master_config_path}")
        self.master_config = mc.config

    def reload_master_config(self):
        """
        If the master config path is set, force reloading
        """
        if self._master_config_path is not None:
            mcp = self.master_config_path
            self.master_config_path = mcp



""" Module containing GettableParameter class """

import time
import logging
import numpy as np
import matplotlib.pyplot as plt
from qcodes.parameters import ParameterWithSetpoints

class GettableParameter(ParameterWithSetpoints):
    """
    This is a valid Gettable not because of inheritance, but because it has the
    expected attributes and methods.
    TODO: zip all streams together, stream at once and unpack while opx resume
    TODO: interleaved measurements

    Attributes:
        unit (str): Unit of the parameter
        label (label): Label for parameter (printed on axis)
        readout (Readout): `Readout` instance that created this parameter
        sequence (Sequence): `Sequence` managing the sequence of QUA program
        qm_job (RunningQmJob): Running job on the opx to interact with
        buffer (obj): QM buffer to get results from
        buffer_val (array): flat numpy array containing results of opx fetch
        shape (tuple): Shape of the setpoints array
        batch_size (tuple): Shape of one OPX batch
        count (int): Amount of successful `get` executions
    """
    def __init__(self, name, sequence, *args, **kwargs) -> None:
        """
        Constructor class for ReadSequence class
        Args:
            name (dict): name of the GettableParameter
            readout (Readout): Readout class summarizing data streams and variables
        """
        super().__init__(name, *args, **kwargs)
        self.unit = ""
        self.label = ""

        self.sequence = sequence
        self.reset_measuerement_attributes()

    def set_raw(self, *args, **kwargs) -> None:
        """Empty abstract `set_raw` method. Parameter not meant to be set"""
        raise NotImplementedError("GettableParameters are not meant to be set")

    def reset_measuerement_attributes(self):
        """Resets all job specific attributes"""
        self.qm_job = None
        self.buffer = None
        self.buffer_val = None
        self.shape = None
        self.batch_size = 0
        self.result_nr = 0
        self.batch_counter = None
        self.nr_registered_results = 0

    def get_raw(self, progress_bar = None) -> np.ndarray:
        """ 
        Get method to retrieve a single batch of data from a running measurement
        On its first call, the gettables attributes get configured regarding
        the given measurement and the underlying hardware.
        """
        logging.debug("GettableParameter %s get_raw called", self.name)
        ### Setup is called on first get
        if self.buffer is None:
            self._set_up_gettable_from_program()

        ### The progress is being tracked while waiting for the buffer to fill
        self._wait_until_buffer_full(progress_bar = progress_bar)
        self.buffer_val = self._fetch_opx_buffer()

        ### The QM can have a delay in populating the stream for big sweeps
        while not self.buffer_val.shape == (self.sequence.sweep_size,):
            time.sleep(0.1)
            self.buffer_val = self._fetch_opx_buffer()
        return self._reshape_data(self.buffer_val, self.shape, self.snaked)

    def _reshape_data(self, a_in: np.ndarray, sizes: tuple[int, ...], snaked: tuple[bool, ...]) -> np.ndarray:
        """
        Reshape the inherited array to be len(sizes). Iterate through each of the sizes gradually
        reshaping the inherited array one dimensions at a time. If for dimension n, snaked[n] is true,
        then reverse every second row of that dimension, then iterate to the next dimension.

        Args:
            a_in (np.ndarray): The input array to be reshaped.
            sizes (Tuple[int, ...]): The target shape dimensions.
            snaked (Tuple[bool, ...]): A tuple indicating which dimensions should have every second row reversed.

        Returns:
            np.ndarray: The reshaped array with specified dimensions and reversed rows as needed
        """

        # First check that the cumulative product of sizes is the same as self.size
        if np.prod(sizes) != a_in.size:
            raise ValueError(f"""The cumulative product of sizes {np.prod(sizes)} must be equal to the 
                             total number of elements in the array {self.size}.""")

        # Duplicate the ndarray
        a = a_in

        # Now iterate through each of the sizes
        for i, (sz, sn) in enumerate(zip(sizes[:-1], snaked[1:])):
            # Reshape a to have an extra dimension of size sz
            if i < len(sizes)-1:
                new_shape = a.shape[:-1] + (sz, -1)
                a = a.reshape(new_shape)

                # If sn is true, then reverse every second innermost row
                if sn:
                    if i == 0: # For the first dimension, reverse elements of even rows
                        a[1::2] = np.flip(a[1::2], axis=-1)
                    else:      # For subsequent dimensions, reverse every second innermost row (like before)
                        a[:, 1::2] = np.flip(a[:, 1::2], axis=-1)
        return a

    def _set_up_gettable_from_program(self):
        """
        Set up Gettable attributes from running OPX including measurement and 
        hardware specific values.
        """
        self.sequence = self.sequence.measurement
        if not self.sequence.measurement.driver.opx:
            raise LookupError("Results cant be retreived without OPX connected")
        self.qm_job = self.sequence.driver.qm_job
        self.batch_counter = getattr(
            self.qm_job.result_handles,
            f"{self.sequence.name}_shots"
        )
        self.buffer = getattr(self.qm_job.result_handles, f"{self.name}_buffer")
        if self.buffer is None:
            raise LookupError(
                f"Buffer {self.name}_buffer not found. Try one of:"
                f"{self.qm_job.result_handles.keys()}")
        self.shape = tuple(reversed(tuple(sweep.length for sweep in self.sequence.sweeps)))
        self.snaked = tuple(reversed(tuple(sweep.snake_scan for sweep in self.sequence.sweeps)))
        self.batch_size = self.sequence.sweep_size

    def _wait_until_buffer_full(self, progress_bar = None):
        """
        This function is running until a batch with self.batch_size is ready
        """
        batch_count, old_count = 0, 0
        last_result_nr = self.result_nr
        time_per_shot = 0
        shot_timing = "Calculate timing...\n"
        total_results = "Total results: ..."
        t0 = time.time()
        try:
            # self.qm_job.resume()
            while batch_count < self.batch_size or not self.qm_job.is_paused():
                logging.debug(
                    "Waiting for buffer to fill (%s/%s), %s",
                    batch_count, self.batch_size, self.qm_job.is_paused()
                    )
                shot_count_result = self.batch_counter.fetch_all()
                if shot_count_result is not None:
                    batch_count = shot_count_result[0]
                    total_nr_results = batch_count + self.nr_registered_results
                    total_results = f"Total results: {total_nr_results}"
                if progress_bar is not None:
                    bar_title = "[cyan]Batch progress "
                    bar_title += f"{batch_count}/{self.batch_size}\n"
                    if batch_count > 0:
                        time_per_shot = 1e3*(time.time()-t0)/(batch_count)
                    shot_timing = f"{time_per_shot:.1f} ms per shot\n"
                    progress_bar[1].update(
                        progress_bar[0],
                        completed = batch_count,
                        description = bar_title + shot_timing + total_results
                    )
                    progress_bar[1].refresh()
        except KeyboardInterrupt as exc:
            raise KeyboardInterrupt('Measurement interrupted by user') from exc
        if progress_bar is not None:
            progress_bar[1].update(progress_bar[0], completed = batch_count)
        self.nr_registered_results += self.batch_size

    def _fetch_opx_buffer(self):
        """
        Fetches the OPX buffer into the `buffer_val` and increments the internal
        counter `count`
        """
        buffer_val = np.array(self.buffer.fetch_all(), dtype = float)
        return buffer_val

    def reset(self):
        """Resets all job specific attributes"""
        self.qm_job = None
        self.buffer = None
        self.buffer_val = None
        self.shape = None
        self.batch_size = 0
        self.result_nr = 0

    def plot_set_current_histogram(self, *args, **kwargs) -> tuple:
        """
        Plots a histrogram of all the measured data so far
        
        Args:
            *args: Arguments for plt.hist
            **kwargs: Keyword arguments for plt.hist
        """
        set_current = np.array(self.get_all(), dtype = float)
        return plt.hist(set_current, args, kwargs)



"""Module containing abstract class for experiments"""
from abc import ABC

class Experiment(ABC):
    """
    Abstract class describing a type of experiment to run on the QM.
    This enforces a certain order of sequences that only requires certain
    input parameters or configurations to be set.
    E.g. for a rabi experiment, the order of sequences is:
        1. Initialization of quantum state
        2. Rabi pulse
        3. Readout of quantum state
    
    Attributes:
        name (str): Name of the experiment to run
        sequences (dict): Sequences to be run within program uploaded to the QM
    """
    _sequences_config = None
    _name = None

    @property
    def name(self) -> str:
        """Name of the experiment to run"""
        return self._name

    @property
    def sequences(self) -> dict:
        """Sequences to be run within program uploaded to the OPX"""
        return self._sequences_config

    def __call__(self) -> dict:
        """Returns the sequences to be run"""
        return self.sequences



""" Helper tools for running and measuring OPX sequences"""
import copy
import time
import logging

import numpy as np
from rich.progress import Progress
from qcodes.dataset import Measurement
from qcodes.dataset.measurements import Runner

def create_measurement_loop(
    sequence,
    measurement: Measurement,
    sweep_list: list[dict],
    register_all: bool = False,
    ):
    """
    Decorator to create a measurement loop for a given measurement. Registers
    all measurement parameters and creates the measurement loops recuresively.
    The OPX measurement can be regarded as the innermost loops of the
    measurement. However the OPX returns all results it measures in the loops in
    the qua measurement in one batch. Therefore in QCoDeS logic a get() on the
    OPX is a single shot measurement. The result is saved on a
    `qcodes.ParameterWithSetpoints` (Matrix with coordinates).

    Loops in the sense of QCoDeS are only added when we sweep parameters outside
    the OPX (e.g) DC voltages LO frequency etc.

    Those loops are added by providing a sweep list. Each entry of dict creates
    a new sweep axis. Parameters (keys) and their setpoints (values ) in the
    respective sweep dict are swept concurrently.

    Args:
        sequence (arbok_driver.Sequence) : The OPX measurement sequence
        measurement (qcodes.Measurement): QCoDeS measurement to run
        sweep list (list[dict[arbok_driver.SequenceParameter, numpy.array]]):
            list configuring the sweep axes and the parameters and setpoints
        register_all (bool, optional): Whether all concurrently swept parameters
            are registered in the measurement. Breaks live plotting.
            Defaults to False

    Returns:
        qcodes.Dataset: Dataset of the measurement
    """
    logging.debug("Creating measurement loop")
    sweep_lengths = [len(next(iter(dic.values()))) for dic in sweep_list]
    nr_sweep_list_points = np.prod(sweep_lengths)
    def decorator(func):
        """Decorator to be returned"""
        def wrapper(*args, **kwargs):
            """Wrapper function with arguments"""
            ### Firstly, all settables are extracted from the sweep dict and the
            ### results arguments are created. Those will used for `add_result`
            result_args_dict = _get_result_arguments(sweep_list, register_all)

            ### The extracted settables are registered in the measurement
            for param, _ in result_args_dict.items():
                logging.debug(
                    "Registering sequence parameter %s", param.full_name)
                measurement.register_parameter(param)

            ### The gettables are registered in the measurement
            for gettable in sequence.gettables:
                gettable_setpoints = result_args_dict.keys()
                logging.debug("Registering gettable %s", gettable_setpoints)
                measurement.register_parameter(
                    gettable, setpoints = gettable_setpoints)

            ### The measurement is run with the recursive measurement loop over
            ### the qcodes (non-opx) parameters
            with measurement.run() as datasaver:
                ### Measurement loops are generated recursively
                with Progress() as progress_tracker:
                    progress_bars = {}

                    progress_bars['total_progress'] = progress_tracker.add_task(
                        description = "[green]Total progress...",
                        total = nr_sweep_list_points)
                    progress_bars['batch_progress'] = progress_tracker.add_task(
                        description = "[cyan]Batch progress...",
                        total = sequence.sweep_size)
                    _create_recursive_measurement_loop(
                        sequence = sequence,
                        datasaver = datasaver,
                        sweeps_list_temp = sweep_list,
                        res_args_dict= result_args_dict,
                        inner_function=func,
                        progress_bars = progress_bars,
                        progress_tracker = progress_tracker,
                        **kwargs
                        )
                    print("Measurement finished!")
                dataset = datasaver.dataset
            return dataset
        return wrapper
    return decorator

def run_arbok_measurement(
    sequence,
    measurement: Measurement,
    sweep_list: list[dict],
    register_all: bool = False,
    ):
    """
    Function calling the decorator `create_measurement_loop` without a function
    """
    filled_decorator = create_measurement_loop(
        sequence = sequence,
        measurement = measurement,
        sweep_list = sweep_list,
        register_all = register_all
        )

    @filled_decorator
    def dummy_function():
        pass
    dataset = dummy_function()
    return dataset

def _get_result_arguments(
        sweep_list: list[dict],
        register_all: bool = False) -> dict:
    """
    Generates a list of parameters that are varied in the sweep and a dict
    
    Args:
        sweep_list (list[dict]): List of dictionairies with params as keys and
            np.ndarrays as settables. Each list entry creates one sweep axis.
            If you want to sweep params concurrently enter more entries into 
            their sweep dict
        register_all (bool): If True, all settables will be registered in the
            measurement. If False, only the first settable of each axis will be
            registered
    Returns:
        result_args_dict (dict): Dict with parameters as keys and tuples of
            (parameter, value) as values. The tuples will be used for
            `add_result` in the measurement
    """
    gettable_setpoints = []
    result_args_dict = {}
    for i, sweep_dict in enumerate(sweep_list):
        for j, param in enumerate(list(sweep_dict.keys())):
            if j == 0 or register_all:
                result_args_dict[param] = ()
                gettable_setpoints.append(param)
            else:
                logging.debug(
                    "Not adding settable %s on axis %s", param.name, i)
    return result_args_dict

def _create_recursive_measurement_loop(
        sequence,
        datasaver: Runner,
        sweeps_list_temp: list,
        res_args_dict: dict,
        progress_bars: list,
        progress_tracker: any,
        *args: any,
        inner_function = None,
        **kwargs: any
        ):
    """
    Recursive generation of parameter sweeps for all axis given in dict
    ext_settables

    Args:
        sweeps_list_temp (list): List of dictionairies of given sweeps
        result_args_temp (list): List of tuples of params their values
    """
    if not sweeps_list_temp:
        ### This is the end of the recursion.
        ### If an inner_function is given it is executed HERE
        if inner_function is not None:
            logging.debug("calling inner function")
            inner_function(*args, **kwargs)

        ### Program is resumed and all gettables are fetched when ready
        sequence.driver.qm_job.resume()
        logging.debug("Job resumed, Fetching gettables")
        result_args_temp = []
        for gettable in sequence.gettables:
            result_args_temp.append(
                (
                    gettable,
                    gettable.get_raw(
                        progress_bar = (
                            progress_bars['batch_progress'], progress_tracker,
                        )
                    )
                )
            )

        ### Retreived results are added to the datasaver
        result_args_temp += list(res_args_dict.values())
        datasaver.add_result(*result_args_temp)
        progress_tracker.update(progress_bars['total_progress'], advance=1)
        progress_tracker.refresh()
        return

    ### The first axis will be popped from the list and iterated over
    sweeps_list_temp = copy.copy(sweeps_list_temp)
    sweep_dict = sweeps_list_temp.pop(0)
    for idx in range(len(list(sweep_dict.values())[0])):
        ### The parameter values are set for the current iteration
        for param, values in sweep_dict.items():
            value = values[idx]
            logging.debug('Setting %s to %s', param.instrument.name, value)
            param.set(value)

            ### The parameter is added to the result arguments dict if its
            ### dict entry is registered
            if param in res_args_dict:
                res_args_dict[param] = (param, value)
            else:
                logging.debug( "Param %s on %s not registered",
                    param.instrument, param.name)
        _create_recursive_measurement_loop(
                *args,
                sequence = sequence,
                datasaver = datasaver,
                sweeps_list_temp = sweeps_list_temp,
                res_args_dict = res_args_dict,
                inner_function = inner_function,
                progress_bars = progress_bars,
                progress_tracker = progress_tracker,
                **kwargs
                )
